#ifdef RCSID
static char RCSid[] =
"$Header: d:/cvsroot/tads/html/htmlpng.cpp,v 1.1.1.1 1999/05/09 21:32:27 MJRoberts Exp $";
#endif

/* Copyright (c) 1997 by Michael J. Roberts.  All Rights Reserved. */
/*
Name
  htmlpng.cpp - PNG image object implementation
Function
  
Notes
  
Modified
  11/08/97 MJRoberts  - Creation
*/

/* include main PNG header */
#include <png.h>

/* include our extended PNG library header */
#include "pngext.h"

/* include TADS OS headers - we need to use TADS OS-layer I/O routines */
#include <os.h>

#ifndef TADSHTML_H
#include "tadshtml.h"
#endif
#ifndef HTML_OS_H
#include "html_os.h"
#endif
#ifndef HTMLPNG_H
#include "htmlpng.h"
#endif


/* ------------------------------------------------------------------------ */
/*
 *   CHtmlPng implementation 
 */

CHtmlPng::CHtmlPng()
{
    /* no image allocated yet */
    rows_ = 0;
    height_ = 0;
    width_ = 0;
    palette_ = 0;
    palette_size_ = 0;
}

CHtmlPng::~CHtmlPng()
{
    /* delete the image */
    delete_image();
}

/*
 *   Delete the existing image 
 */
void CHtmlPng::delete_image()
{
    unsigned long i;

    /* if we have an image, delete it */
    if (rows_ != 0)
    {
        png_bytep *rowp;
        
        /* delete each row */
        for (i = 0, rowp = rows_ ; i < height_ ; ++i, ++rowp)
            th_free(*rowp);

        /* delete the row pointer array */
        th_free(rows_);

        /* no more image */
        rows_ = 0;
        height_ = 0;
    }

    /* if we have a palette, free it */
    if (palette_ != 0)
    {
        delete [] palette_;
        palette_ = 0;
    }
}

/*
 *   Load a PNG file 
 */
int CHtmlPng::read_png_file(const textchar_t *filename,
                            unsigned long seekpos, unsigned long /*filesize*/,
                            int convert_256_color)
{
    png_structp png_ptr;
    png_infop info_ptr;
    png_uint_32 width, height;
    int bit_depth, color_type, interlace_type;
    osfildef *infile;
    double screen_gamma, image_gamma;
    png_color_16 image_background;
    png_color_16 *image_background_ptr;

    /* open the file; return error if it fails */
    if ((infile = osfoprb(filename, OSFTBIN)) == 0)
        return 1;

    /* seek to the start of the image in the file */
    osfseek(infile, seekpos, OSFSK_SET);

    /* delete any existing image */
    delete_image();

    /* create and initialize the png_struct */
    png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING,
                                     (void *)0, (png_error_ptr)0,
                                     (png_error_ptr)0);

    /* if that failed, return an error */
    if (png_ptr == 0)
    {
        osfcls(infile);
        return 1;
    }

    /* allocate and initialize image header memory */
    info_ptr = png_create_info_struct(png_ptr);
    if (info_ptr == 0)
    {
        osfcls(infile);
        png_destroy_read_struct(&png_ptr, (png_infopp)0, (png_infopp)0);
        return 1;
    }

    /* set up error handling */
    if (setjmp(png_ptr->jmpbuf))
    {
        /* an error occurred - clean up and return failure */
        osfcls(infile);
        png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)0);
        return 1;
    }

    /* use our own file reader */
    png_set_read_fn(png_ptr, (void *)infile, &fread_cb);

    /* read the header */
    png_read_info(png_ptr, info_ptr);
    png_get_IHDR(png_ptr, info_ptr, &width, &height, &bit_depth,
                 &color_type, &interlace_type, (int *)0, (int *)0);

    /* reduce color precision to 8 bits per color if it's higher */
    png_set_strip_16(png_ptr);

    /* set the screen gamma value */
    screen_gamma = 2.2;
    if (png_get_gAMA(png_ptr, info_ptr, &image_gamma))
        png_set_gamma(png_ptr, screen_gamma, image_gamma);
    else
        png_set_gamma(png_ptr, screen_gamma, 0.45);

    /* reduce to an 8-bit palette if desired */
    if (convert_256_color)
    {
        int num_palette;
        png_colorp palette;

        if (png_get_PLTE(png_ptr, info_ptr, &palette, &num_palette))
        {
            png_uint_16p histogram;

            png_get_hIST(png_ptr, info_ptr, &histogram);
            png_set_dither(png_ptr, palette, num_palette,
                           256, histogram, FALSE);
        }
        else
        {
            /* dither to a rainbow color cube */
            png_set_rainbow_dither(png_ptr);
        }
    }
    else
    {
        /* ask for full RGB expansion */
        png_set_expand(png_ptr);
    }

    /* 
     *   set the background color - check the file for a background color,
     *   and use it if specified; otherwise, use a default of white 
     */
    if (png_get_bKGD(png_ptr, info_ptr, &image_background_ptr))
    {
        /* there's a background color specified in the file -- use it */
        png_set_background(png_ptr, image_background_ptr,
                           PNG_BACKGROUND_GAMMA_FILE, 1, 1.0);
    }
    else
    {
        /* 
         *   there's no background color specified in the file -- use a
         *   default white background 
         */
        image_background.red = 0xff;
        image_background.green = 0xff;
        image_background.blue = 0xff;
        png_set_background(png_ptr, &image_background,
                           PNG_BACKGROUND_GAMMA_SCREEN, 0, 1.0);
    }

    /* read pixels as BGR */
    png_set_bgr(png_ptr);

    /* swap bytes of 16 bit files to least significant byte first */
    png_set_swap(png_ptr);

    /* update the palette */
    png_read_update_info(png_ptr, info_ptr);

    /* allocate memory */
    allocate_rows(png_ptr, info_ptr, height);

    /* remember the width in pixels and bytes */
    width_ = info_ptr->width;
    row_bytes_ = png_get_rowbytes(png_ptr, info_ptr);

    /* read the image */
    png_read_image(png_ptr, rows_);

    /* done reading */
    png_read_end(png_ptr, info_ptr);

    /*
     *   If we want a color palette, and the image has one (it should),
     *   save the color palette for later use 
     */
    if (convert_256_color && png_ptr->palette != 0)
    {
        png_colorp src;
        HTML_color_t *dst;
        unsigned int i;
        
        /* create the palette */
        palette_size_ = png_ptr->num_palette;
        if (palette_size_ > 256) palette_size_ = 256;
        palette_ = new HTML_color_t[palette_size_];

        /* store it */
        for (dst = palette_, src = png_ptr->palette, i = 0 ;
             i < palette_size_ ; ++i, ++src, ++dst)
            *dst = HTML_make_color(src->red, src->green, src->blue);
    }

    /* done - clean up */
    png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)0);
    osfcls(infile);

    /* success */
    return 0;
}

/*
 *   allocate memory for the image 
 */
void CHtmlPng::allocate_rows(png_struct *png_ptr,
                             png_info *info_ptr,
                             unsigned long height)
{
    png_bytep *rowp;
    png_uint_32 row;
    png_uint_32 rowbytes;

    /* remember the number of rows */
    height_ = height;

    /* allocate the row pointer */
    rows_ = (png_bytep *)th_malloc(height * sizeof(*rows_));

    /* allocate the rows */
    rowbytes = png_get_rowbytes(png_ptr, info_ptr);
    for (row = 0, rowp = rows_ ; row < height ; ++row, ++rowp)
        *rowp = (png_bytep)th_malloc(rowbytes);
}

/*
 *   fread() replacement - we don't use stdio, but instead use the TADS
 *   OS-layer file I/O routines 
 */
void CHtmlPng::fread_cb(png_structp png_ptr, png_bytep buf, png_size_t siz)
{
    osfildef *fp;

    /* get the I/O pointer from png - it's the osfildef pointer */
    fp = (osfildef *)png_get_io_ptr(png_ptr);

    /* read the buffer */
    if (osfrb(fp, buf, siz))
        png_error(png_ptr, "error reading file");
}

